#include "WatchFace.h"
#include "WatchFacePages/Clock.h"

#include <cmath>

RTC_DATA_ATTR bool WatchFace::m_inMenu = false;
RTC_DATA_ATTR Menu WatchFace::m_menu;
RTC_DATA_ATTR uint8_t WatchFace::m_watchFacePage;

void WatchFace::InitBoot()
{
  m_menu.Init(m_display);
  SetupVolatileMenuStuff();
  m_menu.Reset();

  ConnectWiFi();
  SyncNTPTime();
  DisconnectWiFi();

  for (auto & page : m_pages) {
    page->InitBoot();
  }
}

void WatchFace::InitWake()
{
  SetupVolatileMenuStuff();

  for (auto & page : m_pages) {
    page->InitWake();
  }
}

void WatchFace::HandleButtonPress(uint64_t buttonMask)
{
  if (m_inMenu) {
    m_menu.HandleButtonPress(buttonMask);
  } else {
    if (buttonMask & MENU_BTN_MASK) {
      m_inMenu = true;
      m_menu.Redraw(false);

      const std::vector<int> buttons = {MENU_BTN_PIN, UP_BTN_PIN, DOWN_BTN_PIN, BACK_BTN_PIN};
      
      while (m_inMenu) {
        for (int button : buttons) {
          if (digitalRead(button) == HIGH) {
            switch (button) {
              case MENU_BTN_PIN:
                m_menu.HandleButtonPress(MENU_BTN_MASK);
                break;
              case UP_BTN_PIN:
                m_menu.HandleButtonPress(UP_BTN_MASK);
                break;
              case DOWN_BTN_PIN:
                m_menu.HandleButtonPress(DOWN_BTN_MASK);
                break;
              case BACK_BTN_PIN:
                m_menu.HandleButtonPress(BACK_BTN_MASK);
                break;
            }

            while (true) {
              if (digitalRead(button) == LOW) {
                break;
              }

              delay(10);
            }
          }
        }

        delay(10);
      }
    } else if (buttonMask & UP_BTN_MASK) {
      m_watchFacePage = (m_watchFacePage + 1) % m_pages.size();
      ShowWatchFace(true);

      while (true) {
        // Wait for button release
        if (digitalRead(UP_BTN_PIN) == LOW){
          break;
        }

        delay(10);
      }
    } else if (buttonMask & DOWN_BTN_MASK) {
      m_watchFacePage = (m_watchFacePage + m_pages.size() - 1) % m_pages.size();
      ShowWatchFace(true);

      while (true) {
        // Wait for button release
        if (digitalRead(DOWN_BTN_PIN) == LOW){
          break;
        }

        delay(10);
      }
    }
  }
}

void WatchFace::HandleDoubleTap()
{
  ShowWatchFace(true);
}

void WatchFace::HandleTilt()
{
  ShowWatchFace(true);
}

void WatchFace::DrawWatchFace(bool partialRefresh)
{
  if (m_inMenu) {
    return;
  }

  m_pages[m_watchFacePage]->DrawPage(partialRefresh);
}

void WatchFace::SetupVolatileMenuStuff()
{
  static const std::vector<MenuPage> menuPages = {
    {
      0, // backPageNum
      "WATCHY", // title
      ALIGNMENT_CENTER, // titleAlignment
      "", // body
      { // Menu items
        {
          "Sync NTP", // title
          nullptr, // callback
          1 // pageNum
        },
        {
          "Set Timezone", // title
          nullptr, // callback
          2 // pageNum
        },
        {
          "Reset Steps", // title
          nullptr, // callback
          3 // pageNum
        },
        {
          "Back", // title
          nullptr, // callback
          0 // pageNum
        }
      },
    },
    {
      0, // backPageNum
      "SYNC NTP", // title
      ALIGNMENT_CENTER, // titleAlignment
      "Sync with:\n" NTP_SERVER, // body
      { // Menu items
        {
          "Sync", // title
          std::bind(&WatchFace::MenuNTPSyncSelected, this), // callback
          1 // pageNum
        },
        {
          "Back", // title
          nullptr, // callback
          0 // pageNum
        }
      },
    },
    {
      0, // backPageNum
      "TIMEZONE", // title
      ALIGNMENT_CENTER, // titleAlignment
      "", // body
      { // Menu items
        {
          "+0", // title
          std::bind(&WatchFace::MenuTimeZoneSelected, this, 0), // callback
          0 // pageNum
        },
        {
          "+1", // title
          std::bind(&WatchFace::MenuTimeZoneSelected, this, 3600), // callback
          0 // pageNum
        },
        {
          "+2", // title
          std::bind(&WatchFace::MenuTimeZoneSelected, this, 7200), // callback
          0 // pageNum
        },
        {
          "+3", // title
          std::bind(&WatchFace::MenuTimeZoneSelected, this, 10800), // callback
          0 // pageNum
        },
        {
          "Back", // title
          nullptr, // callback
          0 // pageNum
        }
      },
    },
    {
      0, // backPageNum
      "RESET STEPS", // title
      ALIGNMENT_CENTER, // titleAlignment
      "Confirm?", // body
      { // Menu items
        {
          "No", // title
          nullptr, // callback
          0 // pageNum
        },
        {
          "Yes", // title
          std::bind(&WatchFace::MenuConfirmResetSteps, this), // callback
          0 // pageNum
        }
      }
    }
  };

  m_menu.SetPages(menuPages);
  m_menu.SetExitCallback(std::bind(&WatchFace::MenuExited, this));
}

void WatchFace::MenuExited()
{
  if (m_inMenu) {
    m_inMenu = false;
    DrawWatchFace(false);
  }
}

void WatchFace::MenuNTPSyncSelected()
{
  ConnectWiFi();
  SyncNTPTime();
  DisconnectWiFi();
  m_features.rtc.Resync();

  if (m_inMenu) {
    m_inMenu = false;
    m_menu.Reset();
    DrawWatchFace(false);
  }
}

void WatchFace::MenuTimeZoneSelected(int tzOffset)
{
  m_features.storage.SetTzOffset(tzOffset);

  if (m_inMenu) {
    m_inMenu = false;
    m_menu.Reset();
    DrawWatchFace(false);
  }
}

void WatchFace::MenuConfirmResetSteps()
{
  m_features.stepCounter.ResetSteps();

  if (m_inMenu) {
    m_inMenu = false;
    m_menu.Reset();
    DrawWatchFace(false);
  }
}
